#include "main.h"

inline void LoadBgPal(const u16 *src)
{
	Dma3(PAL_BG, src, 256, DMA_DEST_INC | DMA_ENABLE);
}

inline void LoadBgPal16(u8 idx, const u16 *src)
{
	Dma3((void*)(((void*)PAL_BG) + (idx << 5)), (void*)src, 16, DMA_MEMCPY16);
}

inline void LoadObjPal(const u16 *src)
{
	Dma3(PAL_OBJ, src, 256, DMA_DEST_INC | DMA_ENABLE);
}

inline void LoadObjPal16(u8 idx, const u16 *src)
{
	Dma3((void*)(((void*)PAL_OBJ) + (idx << 5)), (void*)src, 16, DMA_MEMCPY16);
}

void memset(void *dest, u8 src, u32 size)
{
	u16 i;

	for(i = 0; i < size; i++)
		*((u8*)dest++) = src;
}

void memcpy(void *dest, void *src, u32 size)
{
	u8 *i;
	for(i = src + size; src < i; src++, dest++)
		*((u8*)dest) = *((u8*)src);
}

inline void Dma0(void *dest, const void *src, u32 count, u16 flags)
{
	REG_DM0SAD = (u32)src;
	REG_DM0DAD = (u32)dest;
	REG_DM0CNT_L = count;
	REG_DM0CNT_H = flags;
}

inline void Dma1(void *dest, const void *src, u32 count, u16 flags)
{
	REG_DM1SAD = (u32)src;
	REG_DM1DAD = (u32)dest;
	REG_DM1CNT_L = count;
	REG_DM1CNT_H = flags;
}

inline void Dma2(void *dest, const void *src, u32 count, u16 flags)
{
	REG_DM2SAD = (u32)src;
	REG_DM2DAD = (u32)dest;
	REG_DM2CNT_L = count;
	REG_DM2CNT_H = flags;
}

inline void Dma3(void *dest, const void *src, u32 count, u16 flags)
{
	REG_DM3SAD = (u32)src;
	REG_DM3DAD = (u32)dest;
	REG_DM3CNT_L = count;
	REG_DM3CNT_H = flags;
}

void MakeTrigTable(s32 *table, u32 size)
{
	BgAffineSrc *bgSrc;
	BgAffineDest *bgDest;
	s32 i;

	bgSrc = eWram;
	eWram += sizeof(BgAffineSrc) << 8;
	bgDest = eWram;
	eWram += sizeof(BgAffineDest) << 8;
	for(i = 0; i < 256; i++)
	{
		bgSrc[i].x = 0;
		bgSrc[i].y = 0;
		bgSrc[i].tX = 0;
		bgSrc[i].tY = 0;
		bgSrc[i].sX = 256;
		bgSrc[i].sY = 256;
		bgSrc[i].theta = i << 8;
	}
	BgAffineSet(bgSrc, bgDest, 256);
	for(i = 0; i < 256; i++)
		table[i] = bgDest[i].pa << 8;
	eWram -= (sizeof(BgAffineSrc) + sizeof(BgAffineDest)) << 8;
}

void MakeRcpTable(s32 *table, u32 size)
{
	s32 i;
	table[0] = 1 << 30;	//technically, this should be infinity, but because Rcp uses linear interpolation, anything below 1 would go up way too fast
	for(i = 1; i < size; i++)
		table[i] = div(1 << 30, i);
}
