#include "main.h"

extern void Snd0EndNoLoop();
extern void Snd0EndLoop();
extern void Snd1EndNoLoop();
extern void Snd1EndLoop();
extern void Mod01EndNoLoop();
extern void Mod01EndLoop();
extern void Mod23EndNoLoop();
extern void Mod23EndLoop();
extern void Mod0EndLoop();
extern void Mod1EndLoop();
extern void Mod2EndLoop();
extern void Mod3EndLoop();

const u32 sndEndTable[] = {
	(u32)Mod01EndNoLoop, (u32)Mod01EndNoLoop, 
	(u32)Mod23EndNoLoop, (u32)Mod23EndNoLoop, 
	(u32)Snd0EndNoLoop, (u32)Snd1EndNoLoop, 
	(u32)Mod0EndLoop, (u32)Mod1EndLoop, 
	(u32)Mod2EndLoop, (u32)Mod3EndLoop, 
	(u32)Snd0EndLoop, (u32)Snd1EndLoop
};

s8 *sndMixBuf, *sndCurMixBuf;
u32 sndMixBufSize, sndFreq;
ModChannel modChn[4];
SndChannel sndChn[2];
Song *pSndSong;

void SndInit(void *mixBuf, void *song)
{
	u32 i;

	REG_SGCNT0_H = SOUNDA_LOUT | SOUNDB_ROUT | 
		SOUNDA_FIFORESET | SOUNDB_FIFORESET | 
		SOUNDA_OUTPUTRATIO | SOUNDB_OUTPUTRATIO;
	REG_SGCNT1 = SOUND_ENABLE;

	sndMixBuf = mixBuf;
	sndMixBufSize = SND_MIXBUFSIZE;
	sndFreq = SND_FREQ;
	pSndSong = (pSong)song;
	pSndSong->state = SND_STATE_BUF;	//so first call to SndVSync sets up DMA
	pSndSong->chnActive = 0;
	pSndSong->vol = 128;
	for(i = 0; i < 4; i++)
	{
		modChn[i].data = dSndBlank;
		modChn[i].pos = 0;
		modChn[i].inc = 0;
		modChn[i].vol = 0;
		modChn[i].len = (SND_MIXBUFSIZE + 1) << SND_FP;
	}
	for(i = 0; i < 2; i++)
	{
		sndChn[i].data = dSndBlank;
		sndChn[i].pos = 0;
		sndChn[i].rVol = 0;
		sndChn[i].lVol = 0;
		sndChn[i].len = SND_MIXBUFSIZE + 1;
	}

	REG_DM1DAD = (u32)(&REG_SGFIFOA);
	REG_DM1CNT_L = 0;
	REG_DM2DAD = (u32)(&REG_SGFIFOB);
	REG_DM2CNT_L = 0;
	REG_TM0D = SND_TMD;
	REG_TM0CNT = TIMER_ENABLE;
}

void SndPlay(u8 *data, u32 len, u16 loopLen, u16 vol, u8 chn)
{
	s32 i;
	u32 flags;

	flags = 0x10;
	if(loopLen != 0xffff)
		flags |= 0x1000;
	if(chn & BIT07)
		flags |= 0x100000;

	if((chn & 0xf) == 0xf)
	{
		chn |= BIT06;
		for(i = 1; i >= 0; i--)
			if(!(pSndSong->chnActive & (0x10 << i)))
				chn = i;
		if(chn & BIT06)
			for(i = 1; i >= 0; i--)
				if(!(pSndSong->chnActive & (0x100000 << i)))
					chn = i;
		if(chn & BIT06)
		{
			if(chn & BIT07)
				chn = 1;
			else
				return;
		}
	}
	pSndSong->chnActive |= flags << chn;
	sndChn[chn].data = data;
	sndChn[chn].pos = 0;
	sndChn[chn].len = len;
	sndChn[chn].loopLen = loopLen;
	*((u16*)(&(sndChn[chn].lVol))) = vol;
}

void SndModChnPlay(u8 *data, u32 inc, u32 len, 
	u32 loopLen, u8 vol, u8 chn)
{
	s32 i;
	u32 flags;

	flags = 1;
	if(loopLen != 0xffffffff)
		flags |= 0x100;
	if(chn & BIT07)
		flags |= 0x10000;

	if((chn & 0xf) == 0xf)
	{
		chn |= BIT06;
		for(i = 1; i >= 0; i--)
			if(!(pSndSong->chnActive & (1 << i)))
				chn = i;
		if(chn & BIT06)
			for(i = 1; i >= 0; i--)
				if(!(pSndSong->chnActive & (0x10000 << i)))
					chn = i;
		if(chn & BIT06)
		{
			if(chn & BIT07)
				chn = 3;
			else
				return;
		}
	}
	pSndSong->chnActive |= flags << chn;
	modChn[chn].data = data;
	modChn[chn].inc = inc;
	modChn[chn].vol = vol;
	modChn[chn].pos = 0;
	modChn[chn].len = len << SND_FP;
	modChn[chn].loopLen = loopLen << SND_FP;
}

u32 SndCalcInc(u32 inc, u32 freq)
{
	return div((inc << SND_FP) * freq, sndFreq << 8);
}

/*
// this is pretty much a direct ASM-to-C conversion of the mixer 
// that I wrote for reference when I gave up on hunting ASM bugs
void SndMix()
{
	//sndArea map: 
	//0-3=order, 4-5=snd vol for this side, 6-7=current mod chns, 
	//8-15=chn dist, 16-23=sample count stack, 
	//24=loop, 25=sample count stack idx, 28-31=end jump table, 
	//32-35=signAdjust, 36-39=blankData, 
	//40-43=modChn, 44-47=sndChn, 48-51=r13, 52-55=mixBuf, 
	//56-59=SndMixLoop, 60-63=return from SndMixLoop
	s32 i, j, k, l;
	u32 sndArea[16];
	u8 *chnOrder = ((u8*)sndArea), 
		*sndVol = ((u8*)sndArea) + 4, 
		*mod = ((u8*)sndArea) + 6, sndEnd;
	s16 *chnDist = ((s16*)sndArea) + 4, 
		*sampCnt = ((s16*)sndArea) + 8;

	*((u32*)0x3000000) = (u32)sndArea;
	((u32*)sndArea)[7] = (u32)sndEndTable;
	((u32*)sndArea)[9] = (u32)dSndBlank;
	((u32*)sndArea)[10] = (u32)modChn;
	((u32*)sndArea)[11] = (u32)sndChn;
	((u32*)sndArea)[13] = (u32)sndCurMixBuf;
	mod[0] = 0;
	mod[1] = 3;
	sndVol[0] = sndChn[0].lVol;
	sndVol[1] = sndChn[1].lVol;
	for(k = 2; k != 0; k--)
	{
		chnOrder[0] = mod[0];
		chnOrder[1] = mod[1];
		chnOrder[2] = 4;
		chnOrder[3] = 5;

		((s32*)sndArea)[8] = 0;	//sign adjust
		sndEnd = 0;
		for(i = 1; i >= 0; i--)
		{
			if((modChn[mod[i]].pos + modChn[mod[i]].inc * 
				SND_MIXBUFSIZE) < modChn[mod[i]].len)
				chnDist[i] = SND_MIXBUFSIZE;
			else
			{
				chnDist[i] = div(modChn[mod[i]].len - 
					modChn[mod[i]].pos, modChn[mod[i]].inc);
				sndEnd |= 1 << mod[i];
			}
			((s32*)sndArea)[8] -= modChn[mod[i]].vol << 7;
		}
		for(i = 1; i >= 0; i--)
		{
			if((sndChn[i].pos + SND_MIXBUFSIZE) < sndChn[i].len)
				chnDist[i + 2] = SND_MIXBUFSIZE;
			else
			{
				chnDist[i + 2] = sndChn[i].len - sndChn[i].pos;
				sndEnd |= 16 << i;
			}
			((s32*)sndArea)[8] -= sndVol[i] << 7;
		}
		for(j = 0; j < 3; j++)
		{
			l = j;
			for(i = j + 1; i < 4; i++)
				if(chnDist[i] > chnDist[l])
					l = i;
			if(l != j)
			{
				chnDist[l] += chnDist[j];
				chnDist[j] = chnDist[l] - chnDist[j];
				chnDist[l] -= chnDist[j];
				chnOrder[l] += chnOrder[j];
				chnOrder[j] = chnOrder[l] - chnOrder[j];
				chnOrder[l] -= chnOrder[j];
			}
		}
		sampCnt[3] = chnDist[3];
		for(i = 2; i >= 0; i--)
			sampCnt[i] = chnDist[i] - chnDist[i + 1];
		((u8*)sndArea)[24] = 0;
		for(i = 3; i >= 0; i--)
			if(pSndSong->chnActive & (0x100 << chnOrder[i]))
				((u8*)sndArea)[24] |= i << i;

		((u8*)sndArea)[25] = 4;	//chn end stack counter
		m0p = modChn[mod[0]].pos + (((u32)(modChn[mod[0]].data)) << 8);
		m0i = modChn[mod[0]].inc;
		m0v = modChn[mod[0]].vol;
		m1p = modChn[mod[1]].pos + (((u32)(modChn[mod[1]].data)) << 8);
		m1i = modChn[mod[1]].inc;
		m1v = modChn[mod[1]].vol;
		s0p = (u32)(sndChn[0].data + sndChn[0].pos);
		s0v = sndVol[0];
		s1p = (u32)(sndChn[1].data + sndChn[1].pos);
		s1v = sndVol[1];
		mBuf = (s8*)(((u32*)sndArea)[13]);
		sgnAdj = ((s32*)sndArea)[8];
		while(1)
		{
			((s8*)sndArea)[25] -= 1;
			if(((s8*)sndArea)[25] < 0)
			{
				if(chnDist[0] < SND_MIXBUFSIZE)
				{
					sCnt = SND_MIXBUFSIZE - chnDist[0];
					while(sCnt != 0)
					{
						*mBuf++ = 0;
						sCnt--;
					} 
				}
				goto snd_end;
			}
			else if(sampCnt[((u8*)sndArea)[25]] > 0)
			{
				sCnt = sampCnt[((u8*)sndArea)[25]];
				__FarProcedure((u32)SndMixLoop);
			}
			u32 tmp = chnOrder[((u8*)sndArea)[25]];
			if(sndEnd & (1 << tmp))
			{
				if(((u8*)sndArea)[24] & (1 << ((u8*)sndArea)[25]))
					tmp += 6;
				__FarProcedure(sndEndTable[tmp]);
			}
		}
snd_end:;

		for(i = 3; i >= 0; i--)
		{
			j = chnOrder[i];
			if(j < 4)
			{
				if(pSndSong->chnActive & (1 << j))
				{
					if(sndEnd & (1 << j))
					{
						if(pSndSong->chnActive & (0x100 << j))
							modChn[j].pos += (modChn[j].inc * SND_MIXBUFSIZE) - 
								modChn[j].loopLen;
						else
						{
							modChn[j].data = dSndBlank;
							modChn[j].pos = 0;
							modChn[j].inc = 256;
							modChn[j].len = (SND_MIXBUFSIZE + 1) << 8;
							modChn[j].vol = 0;
							pSndSong->chnActive &= ~(0x101 << j);
						}
					}
					else
						modChn[j].pos += modChn[j].inc * SND_MIXBUFSIZE;
				}
				else
					modChn[j].pos = 0;
			}
		}
//set up for second pass
		mod[0] = 1;
		mod[1] = 2;
		sndVol[0] = sndChn[0].rVol;
		sndVol[1] = sndChn[1].rVol;
		((u32*)sndArea)[13] = (u32)(sndCurMixBuf + (SND_MIXBUFSIZE << 1));
	}
	for(i = 1; i >= 0; i--)
	{
		if(pSndSong->chnActive & (0x10 << i))
		{
			if(sndChn[i].len - sndChn[i].pos <= SND_MIXBUFSIZE)
			{
				if(pSndSong->chnActive & (0x1000 << i))
					sndChn[i].pos += SND_MIXBUFSIZE - sndChn[i].loopLen;
				else
				{
					sndChn[i].data = dSndBlank;
					sndChn[i].pos = 0;
					sndChn[i].len = SND_MIXBUFSIZE + 1;
					sndChn[i].lVol = 0;
					sndChn[i].rVol = 0;
					pSndSong->chnActive &= ~(0x1010 << i);
				}
			}
			else
				sndChn[i].pos += SND_MIXBUFSIZE;
		}
		else
			sndChn[i].pos = 0;
	}
}

void Snd0EndNoLoop()
{
	s0p = (u32)dSndBlank;
}

void Snd0EndLoop()
{
	s0p -= sndChn[0].loopLen;
}

void Snd1EndNoLoop()
{
	s1p = (u32)dSndBlank;
}

void Snd1EndLoop()
{
	s1p -= sndChn[1].loopLen;
}

void Mod01EndNoLoop()
{
	m0p = ((u32)dSndBlank) << 8;
	m0i = 256;
}

void Mod23EndNoLoop()
{
	m1p = ((u32)dSndBlank) << 8;
	m1i = 256;
}

void Mod0EndLoop()
{
	m0p -= modChn[0].loopLen;
}

void Mod1EndLoop()
{
	m0p -= modChn[1].loopLen;
}

void Mod2EndLoop()
{
	m1p -= modChn[2].loopLen;
}

void Mod3EndLoop()
{
	m1p -= modChn[3].loopLen;
}
*/