#ifndef SOUND_H
#define SOUND_H

#include "main.h"

#define SND_FP 8
#define SND_CHANNELS 4
#define SND_MIXBUFSIZE 304
#define SND_TMD 64612
#define SND_FREQ 18157
//settings for different mixing rates, in the format 
	//SND_TMD = 65536 - (16777216 / SND_FREQ), buf = SND_MIXBUFSIZE
	//62610 = 65536 - (16777216 / 5734), buf = 96
	//63940 = 65536 - (16777216 / 10512), buf = 176
	//64282 = 65536 - (16777216 / 13379), buf = 224
	//64472 = 65536 - (16777216 / 15768), buf = 264
	//64612 = 65536 - (16777216 / 18157), buf = 304
	//65004 = 65536 - (16777216 / 31536), buf = 528
	//65154 = 65536 - (16777216 / 43959), buf = 736

#define SND_STATE_BUF	BIT00
#define SND_STATE_PLAY	BIT02
#define SND_STATE_PAUSE	BIT03

typedef struct tagSample
{
	u16 sndIdx;
	u8 tune, volLoop;	//bit0-6 vol, bit7 loop flag
} Sample, *pSample;

typedef struct tagSong
{
	u32 chnActive;	//byte0=active, byte1=loop, byte2=priority, byte3=effect neg
	u8 vol, state, bpm, spd;
	Sample *samp;
	void *songData;	//song currently playing
	u8 *rowPtr;
	u8 *order;
	u8 *patData;
	u16 *patOffset;
	u16 tick, incTick;
	s8 pos, curRow;
	u8 len, chnCnt, 
		patCnt, sampCnt, patDelay, patLoopRow, 
		patLoopCnt, pad1, pad2, pad3;
	u16 chnPeriod[SND_CHANNELS];
	u8 chnNote[SND_CHANNELS], chnSamp[SND_CHANNELS], 
		chnEffect[SND_CHANNELS], chnParam[SND_CHANNELS], 
		chnVol[SND_CHANNELS], chnPan[SND_CHANNELS], 
		chnPatLoopRow[SND_CHANNELS], chnFxPos[SND_CHANNELS], 
		chnFxFlags[SND_CHANNELS], chnPortaSpd[SND_CHANNELS], 
		chnVibratoSpd[SND_CHANNELS], chnTremoloSpd[SND_CHANNELS], 
		chnFxTemp[SND_CHANNELS];	//fxFlags: bit0-2 vibrato, bit3-5 tremolo, bit7 glissando
} Song, *pSong;

typedef struct tModChannel
{
	u32 data, pos, inc, vol, len, loopLen;	//pos, inc, len and loopStart are 24.8 fp
} ModChannel, *pModChannel;

typedef struct tSndChannel
{
	u32 data, pos, len;
	u16 loopLen;
	u8 lVol, rVol;
} SndChannel, *pSndChannel;

extern s8 *sndMixBuf, *sndCurMixBuf;
extern u32 sndMixBufSize, sndFreq;
extern ModChannel modChn[4];
extern SndChannel sndChn[2];
extern pSong pSndSong;
extern const u16 sndPeriodTable[];
extern const u8 sndSineTable[];

extern void SndInit(void*, void*);	//mix buf data (size SND_MIXBUFSIZE * 4), main song data (size sizeof(Song)), both are normally in EWRAM, minor speedup in IWRAM
extern void SndVSync() IN_IWRAM;		//this doesn't have to be in IWRAM
extern void SndMix();
extern void SndMixLoop() IN_IWRAM;
extern void SndPlay(u8*, u32, u16, u16, u8);	//data. len, loopLen (0xffff for none), vol (bit0-6 l, 8-14 r), chn (0-1, 0xf for auto, bit7=priority)
extern void SndModChnPlay(u8*, u32, u32, u32, u8, u8);	//data. inc, len, loopLen (0xffffffff for none), vol, chn (0-3, 0xf for auto, bit7=priority)
extern u32 SndCalcInc(u32, u32);				//inc (24.8 fp), freq
extern void SndSongPlay(u8*, u8, u8);		//song data, start order, start row
extern void SndSongStop();						//stop song and all low priority channels
extern void SndSongPause(u32);				//0=unpause, 1=pause and stop all low priority channels
extern void SndSongSeek(u8, u8);				//order, row
extern void SndSongUpdate();

#endif
